%--------------------------------------------------------------------------------------------------------------
% DSDLL demo: Simple soundcard crosstalk meter by Stanislav Maslan
%--------------------------------------------------------------------------------------------------------------
% This script measures output->input and input->input crosstalks. It measures both magnitude and
% phase for every specified frequency. For its function it requires a multiplexer, see project WWW for details.
%
% Connection: Second output channel is used to generate sinewave of given frequency. First channel 
% generates positive and negative pulses for switching the relays of the multiplexer. The multiplexer is 
% connected between the second output and one of the inputs. The unused input is terminated by a desired load.
%
% Measurement process for each frequency: The second channel starts generation of the sinewave. First channel
% issues positive pulse to enable the multiplexer path. Then it waits to settle the transition for given time.
% Next it samples both input channels to get U1L and U1R voltages (obtained using FFT). Then it issues 
% negative pulse to disable the multiplexer path. After settling it samples again to get voltages U0L and U0R.
% Both multiplexer steps are repeated 'A' times and the results are averaged so it may take quite some time.
% Output to L/R input crosstalk is calculated as:
%  ct_L = U0L/U1L 
%  ct_R = U0R/U1L
% Channel to channel crosstalk between the input channels is calculated as:
%  ct_LR = U1R/U1L - ct_R
%
% The measured crosstalk will be dependent on the output loading impedance and also on the input termination 
% impedance so a proper termination must be chosen. This script measures the crosstalk as a complex transfer 
% coefficients and the data are stored to the CSV file.
%
% Keep in mind the L/R channels may be swapped on various soundcards so if the system shows 
% some nonsense like positive dB values try to swap the channels. Also the multiplexer relays may be inverted.
%
% The name of the DirectSound devices must be entered in the script! Type in the name as it is
% shown when the script prints out the enumerated devices. Or leave empty so it will choose your 
% system default.
%
% For optimal function set soundcard output/capture mixing rates to the same as in this script i.e. 
% prefereably 48kHz/24bit. Otherwise Windoze will resample the signals.
%
% This demo requires 'dsrec.exe', 'dsdll.dll' files in the pwd (see project WWW).
% It also requires 'ds_run_mpx.m' and 'ds_enum.m' functions in pwd.
%
%
% LICENSE
% -------
% Copyright  2016, Stanislav Maslan <s.maslan@seznam.cz>
%
% This script is free software. It comes without any warranty. You can use it, redistribute it
% and/or modify it as you like as long as this license, the author and link to the 
% DSDLL project web page is included. 
%
% Author: Stanislav Maslan
% E-mail: s.maslan@seznam.cz
% www: http://elektronika.kvalitne.cz/SW/dsdll/dsdll_eng.html
% Revision: V1.0, 2016-08-31
% Tested: Octave 3.6.4+
%--------------------------------------------------------------------------------------------------------------
clear all;
close all;
clc;

% set pwd to this script directory otherwise it won't find the 'dsrec.exe' utility
mfld = [fileparts(mfilename('fullpath')) filesep()];
cd(mfld);

% info
printf('Simple soundcard crosstalk meter by Stanislav Maslan\nusing DirectSound DLL interface DSDLL.\n\n');


% --- general setup
% crosstalk result file
crot_file = 'crot.csv';
% show spectra on runtime
show_spectrum = 1;
% reference input channel (the one connected to the output of the multiplexer) {1 or 2}
ref_in = 1; 


% --- ADC setup
% sampling window size [s]
% use longer period for low frequencies or lower standard deviation of the measurement
T = 3.0;
% multiplexer relay settling time (at least few hundred ms) [s]
% this is essential to prevent transients in the waveform after relay swithc
% use longer values for low frequencies
% use vector parameters to set various settling times for various frequencies, eg.
% settle_f = [20 500 1000 20000]; settle = [5 2 1 0.3];
% the actual settling time will be interpolated from the two vectors 
settle = [2.5 1.5 1 0.7 0.3 0.2];
settle_f = [20 100 200 400 1000 20000];

% spectrum averages count (use at least 5, rather 10+)
A = 2;


% --- DDS setup
% sampling rate [Hz]
fs = 48000;
% desired list of frequencies to synthesize [Hz]
fx = logspace(log10(40),log10(20000),50);
% amplitude [-]
amp = 0.9;


% --- string names of the input and output DirectSound devices, leave empty to use default ones
dsinp_name = 'Radek (USB Sound Blaster HD)';
dsout_name = 'Reproduktory (USB Sound Blaster HD)';



% --- enumerate DirectSound devices, find desired devices IDs
[dsinp_id,dsout_id,dsinp,dsout] = ds_enum(dsinp_name,dsout_name);

% print list of DirectSound devices
printf('Output devices:\n');
for k = 1:numel(dsout), printf('#%d - %s\n',k - 1,dsout{k}); endfor
printf('\nCapture devices:\n');
for k = 1:numel(dsinp), printf('#%d - %s\n',k - 1,dsinp{k}); endfor
printf('\n\n');




% === SWEEP LOOP ===
fz = [];
ct_L = [];
ct_R = [];
ct_LR = [];
for k = 1:numel(fx)

  % frequency of the current sweep step [Hz]
  f = fx(k);
    
  % --- calculate DDS setup so it is coherent with the loop buffer length
  % possible number of periods per the loop buffer  
  M = [1 3 5 7]; 
  % possible samples count for every M 
  N = round(fs./(f./M));
  % possible aligned (coherent) frequencies for every M
  fm = fs./(N./M);
  % select the M for lowest aligned frequency deviation from the desired one
  [v,id] = min(abs(fm - f));
  % get the aligned (coherent) frequency [Hz]
  f = fm(id);
  % get the total number of samples
  N = N(id);
  
  % calculate samples count for recorder so it is coherent with the DDS wave period
  % and so the record is at least 'T' seconds long 
  M = ceil((T*fs)/N)*N;
  
  % finally expand samples count to minimum waveform size for DSDLL,
  % becuase DS don't like too short playback loops
  if N < 16, N *= ceil(16/N); endif
  
  % just in case leave out frequencies which are above nyquist limit
  if(f > 0.45*fs)
    break;
  endif
  
  
  % --- synthesize the sine waveform
  % generate phase vector [rad]
  p = [];
  p(:,1) = [0:N-1]*f/fs*2*pi;    
  % generate the waveform loop for the DDS
  ugen = amp*sin(p);
  
  
  % --- perform the main mesurement
  
  % multiplexing sequence length
  S = 2;
  % build multiplexing sequence: repeat sequence [1: on, 0: off] average 'A' times
  mpx = repmat([1 0],[1 A]);
  
  % get settling time for current frequency
  rel_set = interp1(settle_f,settle,f,'extrap');
    
  % digitize
  [u,ts] = ds_run_mpx(dsinp_id,dsout_id,ugen,fs,2,0,mpx,rel_set,M);
  
  
  % --- process the data
  % calculate spectra of all captured waveforms
  % cut the DC and negative frequencies from the spectra, normalize 
  U = fft(u)(2:floor(size(u,1)/2),:,:)*2/size(u,1);

  % build frequency scale for the spectra
  fc = [];
  fc(:,1) = [1:size(U,1)]/size(u,1)*fs;

  % fix the phases of the particular captured waves by timestamps so they can be treated like they were sampled simultaneously
  % U' = U*exp(-j*w*time_stamp)
  U .*= exp(-j*2*pi*fc.*ts);

  % split the measured sequence per averaging cycles
  U = reshape(U,[size(U,1) size(U,2) S A]);

  % average spectra in complex domain
  if A > 1
    U = mean(U,4);
  endif
  
  % fundamental frequency bin ID
  fid = round(f/(fs/M)); 
    
  % extract the DFT bins with the fundamental frequency f 
  Uf = U(fid,:,:);
  
  % store frequency
  fz(k,1) = f;
  
  if ref_in == 1
    % --- multiplexer connected to 1. channel
    % crosstalk from generator channel to left line-in (1. channel - the one used as a reference) 
    ct_L(k,1) = Uf(1,1,2)./Uf(1,1,1);
    
    % crosstalk from generator channel to right line-in (2. channel - the one terminated with a short)
    ct_R(k,1) = Uf(1,2,2)./Uf(1,1,1);   
     
    % crosstalk from the left line-in to right line-in (1. channel to 2. channel)
    ct_LR(k,1) = Uf(1,2,1)./Uf(1,1,1) - ct_R(k,1);
    
    % display result
    printf('f = %7.1f Hz, out->L = %6.1f dB, out->R = %6.1f dB, L->R = %6.1f dB\n',f,20*log10(abs(ct_L(k))),20*log10(abs(ct_R(k))),20*log10(abs(ct_LR(k))));
    
  else
    % --- multiplexer connected to 2. channel
    % crosstalk from generator channel to left line-in (1. channel - the one terminated with a short) 
    ct_L(k,1) = Uf(1,1,2)./Uf(1,2,1);
    
    % crosstalk from generator channel to right line-in (2. channel - the one used as a reference)
    ct_R(k,1) = Uf(1,2,2)./Uf(1,2,1);   
     
    % crosstalk from the right line-in to left line-in (2. channel to 1. channel)
    ct_LR(k,1) = Uf(1,1,1)./Uf(1,2,1) - ct_L(k,1);
    
    % display result
    printf('f = %7.1f Hz, out->L = %6.1f dB, out->R = %6.1f dB, R->L = %6.1f dB\n',f,20*log10(abs(ct_L(k))),20*log10(abs(ct_R(k))),20*log10(abs(ct_LR(k))));
  
  endif
  
  
  % display spectrum?
  if show_spectrum
    figure(2);
    loglog(fc,abs(U(:,ref_in,1)));
    xlabel('f [Hz]');       
    ylabel('U [V]');
    grid on;
  endif
   
  if numel(kbhit(1))
    % leave if key pressed...    
    break;
  endif 

endfor

% store measured dependence into CSV file
data = [fz real(ct_L) imag(ct_L) real(ct_R) imag(ct_R) real(ct_LR) imag(ct_LR)];
if ref_in == 1
  csv = {'f','re{O-L}','im{O-L}','re{O-R}','im{O-R}','re{L-R}','im{L-R}';'Hz','-','-','-','-','-','-'};
else
  csv = {'f','re{O-L}','im{O-L}','re{O-R}','im{O-R}','re{R-L}','im{R-L}';'Hz','-','-','-','-','-','-'};
endif
for k = 1:numel(fz)
  csv(end+1,:) = num2cell(data(k,:));
endfor
cell2csv(crot_file,csv,';');

% show the measured dependencies
figure(1);
semilogx(fz,20*log10(abs([ct_L ct_R ct_LR])),'LineWidth',1.5);
xlabel('f [Hz]');
ylabel('A [dB]');
box on;
grid on;
if ref_in == 1
  legend('out -> left-in','out -> right-in','left-in -> right-in','location','northwest');
else
  legend('out -> left-in','out -> right-in','right-in -> left-in','location','northwest');
endif  

% print it to an image
%print('crosstalk.png','-S640,400','-F:7','-dpng');