%--------------------------------------------------------------------------------------------------------------
% This script uses DirectSound interface DSDLL in a time multiplexing mode.
% It loops the user waveform 'out_data' at second output channel of the DS output device 'dsout_id'.
% First channel of the output DS device is used to generate positive and negative pulses that
% are used to control the multiplexer relays (just two states are supported).
% For every multiplexer state from the sequence 'mpx_sequence' this function waits to settle for 
% 'mpx_delay' seconds and then samples 'rec_samples' samples from capture DS device 'dsinp_id'. Then it 
% waits 'rec_post' seconds before next multiplexing sequence item is processed.  
% Width 'mpx_pulse' and amplitude 'mpx_amp' of the control pulses can be customized.
% The function also emulates timestamps so for every captured waveform in the 'rec_data' it also returns
% relative timestamp 'time_stamp' of the first captured sample.  
%
% This function requires 'dsrec.exe' and 'dsdll.dll' files in the pwd (see project WWW).
% Transfer of the playback and captured waveform data is made via *.mat files.  
%
% parameters:
%   dsinp_id - capture device ID returned by ds_enum(), use 0 for default capture device
%   dsinp_id - output device ID returned by ds_enum(), use 0 for default output device
%   out_data - waveform data for playback, single column (one channel only!), normalized to <-1;+1>
%            - leave empty to disable playback, control pulses are still generated, but nothing on the
%              the second channel
%   fs - sampling rate (integer), common for capture and output [Hz]
%   rec_chn - number of channels of recorder (1 or 2)
%   rec_delay - delay between start of playback and start of measurement sequence [s]
%   mpx_sequence - multiplexing sequence, i.e. vector of multiplexer states (0 or 1)
%   mpx_delay - multiplexer settling time, can be scalar with a common delay for entire sequence [s]  
%               or vector of the same length as 'mpx_sequence' [s]
%   rec_samples - number of samples to capture for every 'mpx_sequence' item [-] 
%   mpx_pulse - width of the multiplexer control pulse (default 0.001s) [s]
%   mpx_amp - multiplexer control pulse amplitude (default 0.95) [-]
%   res_post - post capture safety delay (default 0.1s) [s]
%   out_volume - output volume in dB (-100 to 0) dB, use 0 by default for maximum volume    
%
% returns:
%   rec_data - 3-dimensional array of captured samples, normalized to <-1;+1>:
%              rows: samples, columns: input channels, slices: 'mpx_sequence' items 
%   time_stamp - 3-dimensional array of timestamps, data organisation is the same as 'rec_data'
%   raw_capture - raw captured waveform for debuging purposes, one column per channel
%   raw_output - raw output waveforms generated at the DS output device, one column per channel  
%
%
% LICENSE
% -------
% Copyright  2016, Stanislav Maslan <s.maslan@seznam.cz>
%
% This script is free software. It comes without any warranty. You can use it, redistribute it
% and/or modify it as you like as long as this license, the author and link to the 
% DSDLL project web page is included. 
%
% Author: Stanislav Maslan
% E-mail: s.maslan@seznam.cz
% www: http://elektronika.kvalitne.cz/SW/dsdll/dsdll_eng.html
% Revision: V1.0, 2016-08-31
% Tested: Octave 3.6.4+
%--------------------------------------------------------------------------------------------------------------
function [rec_data,time_stamp,raw_capture,raw_output] = ds_run_mpx(dsinp_id,dsout_id,out_data,fs,rec_chn,rec_delay,mpx_sequence,mpx_delay,rec_samples,mpx_pulse = 0.001,mpx_amp = 0.95,rec_post = 0.1,out_volume = 0)

  % --- check input parameters 
  
  % fix output data
  out_data = out_data(:);
  
  % check the post capture delay
  if rec_post < 0
    error('Post capture delay ''mpx_post'' must be positive!');
  endif
  rec_post = round(fs*rec_post);
  
  % check multiplexing sequence parameter
  if ~numel(mpx_sequence)
    error('Multiplexing sequence ''mpx_sequence'' is empty!');
  endif
  if any(mpx_sequence > 1) || any(mpx_sequence < 0)
    error('Multiplexing sequence ''mpx_sequence'' can contain only channels 0 or 1!');
  endif  
  mpx_sequence = 2*mpx_amp*(0.5 - mpx_sequence(:));
  
  % check capture length parameter
  if numel(rec_samples) ~= 1
    error('Samples count to capture ''rec_samples'' must be scalar!');
  endif
  if rec_samples < 1
    error('Samples count to capture ''rec_samples'' must be greater than 0!');
  endif
    
  % check multiplexer settling time parameter
  if numel(mpx_delay) > 1 && numel(mpx_delay) ~= numel(mpx_sequence) 
    error('Multiplexer settling time ''mpx_delay'' must be either scalar or the same size as length of multiplexing sequence ''mpx_sequence''!');
  elseif ~numel(mpx_delay)
    error('No multiplexer settling time ''mpx_delay'' specified!');
  elseif numel(mpx_delay) == 1
    % settling time is common for all multiplexing sequence steps, make vector of it
    mpx_delay = repmat(mpx_delay,size(mpx_sequence));
  endif
  if any(mpx_delay < 0)
    error('Multiplexer settling time ''mpx_delay'' must be positive!');
  endif
  % convert the time into samples count
  mpx_delay = round(mpx_delay(:)*fs);
  
  % check initial capture delay parameter
  if ~numel(rec_delay) || rec_delay < 0
    error('Initial capturing delay ''rec_delay'' must be positive!');
  endif  
  
  % check minimum length of a multiplexer latching pulse
  if fs*mpx_pulse < 3
    error('Multiplexer setup pulse ''mpx_pulse'' is too short!');  
  endif
  % convert pulse length to samples
  mpx_pulse = round(fs*mpx_pulse); 
  
  
  
  % --- build multiplexer latching sequence
  
  % multiplex sequence length
  M = numel(mpx_sequence);
  
  % output waveform
  ugen = [];
  
  % particular capture offsets
  caps = [];
  
  % prepend some safety delay to the start of the multiplexing latching sequence
  ugen = [ugen;zeros(round((0.05 + rec_delay)*fs),1)];
  
  % build the sequence data
  for m = 1:M
    % for each multiplexing step
    
    % append latching pulse
    ugen = [ugen;repmat(mpx_sequence(m),mpx_pulse,1)];
    
    % append multiplexer settling delay
    ugen = [ugen;repmat(0,mpx_delay(m),1)];
    
    % remember capture start position
    caps = [caps;numel(ugen) + 1];
    
    % append capture length delay
    ugen = [ugen;repmat(0,rec_samples,1)];
    
    % append capture length delay
    ugen = [ugen;repmat(0,rec_post,1)];
     
  endfor
  
  % append some safety delay to the end of the sequence
  ugen = [ugen;zeros(round(0.2*fs),1)];
  
  % output waveform length
  G = numel(out_data);
  
  % latching waveform length
  N = numel(ugen);
  
  % now add the output waveform to the second channel
  if numel(out_data)
    % output data available, repeat it in loop  
    
    % align total length of the latching sequence to the least multiple of output data length 
    T = ceil(N/G)*G;  
    ugen = resize(ugen,T,1);
    
    % loop the output data to secondary channel
    ugen = [ugen repmat(out_data,ceil(N/G),1)];
  
  else
    % no output data, so generate silence into secondary channel    
    ugen = [ugen zeros(N,1)];
            
  endif
  
  % total output waveform length
  N = size(ugen,1);
    
  
  % save output playback data to temp file
  ofile = '_ds_dll_out.mat';
  save('-v4',ofile,'ugen');
    
  % call the DSDLL interface, play output in the loop because it will end playback when the capture is done
  ss = sprintf('dsrec.exe -p -c%d -s%d -d%d -o%d %s -i%d %d _ds_dll_capt.mat -v%dp0 -l',int32(rec_chn),int32(fs),int32(0),int32(dsout_id),ofile,int32(dsinp_id),N,int32(out_volume*100));
  [ec,stdout] = system(ss);
  
  % remove playback temp file
  unlink('_ds_dll_out.mat');
  
  % check for DSDLL error
  if ec
    error(stdout);
  endif
  
  % try to load captured waveform
  load('-v4','_ds_dll_capt.mat','data');    
    
  % remove captured waveform file
  unlink('_ds_dll_capt.mat');
  
  % captured data
  rec_data = [];
  
  % extract the particular captured waveforms from the digitized noodle (samples,channel,sequence_item) 
  for m = 1:M
    rec_data(:,:,m) = data(caps(m):caps(m)+rec_samples-1,:);
  endfor
  
  % build timestamps for every channel of every sequence item (1,channel,sequence_item) [s]
  time_stamp = reshape(repmat(caps,1,rec_chn).',[1 rec_chn M])/fs;
  
  % return raw captured waveform data
  raw_capture = data;
  
  % return raw output waveform data
  raw_output = ugen;

endfunction